// [[file:../arroyo-native-parser.org::*First Pass][First Pass:1]]
use anyhow::Result;
use regex;
use std::borrow::Cow;
use std::fs;
use std::io::{Error, Write};
use std::marker::PhantomData;

use orgize::export::{DefaultHtmlHandler, HtmlEscape, HtmlHandler, SyntectHtmlHandler};
use orgize::{elements, Element, Org};

use crate::export_html::ArroyoHtmlHandler;
use crate::export_html::ExportOptions;
// First Pass:1 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:2]]
pub struct ArroyoAtomHandler<E: From<Error>, H: HtmlHandler<E>> {
    pub options: ExportOptions,
    pub inner: ArroyoHtmlHandler<E, H>,
    pub error_type: PhantomData<E>,

    // internal parser state
    in_heading: bool,
    in_drawer: bool,
    heading_lvl: usize,

    // Document metadata placed in feed
    pub filetags: Vec<String>,
    pub authors: Vec<String>,
    pub feed_title: String,
    pub feed_page_id: String,
    pub last_date: String,
}

impl<E: From<Error>, H: HtmlHandler<E>> ArroyoAtomHandler<E, H> {
    pub fn new(options: ExportOptions, inner: ArroyoHtmlHandler<E, H>) -> Self {
        ArroyoAtomHandler {
            inner,
            options,
            ..Default::default()
        }
    }
}

impl<E: From<Error>, H: HtmlHandler<E>> Default for ArroyoAtomHandler<E, H> {
    fn default() -> Self {
        ArroyoAtomHandler {
            inner: ArroyoHtmlHandler::default(),
            error_type: PhantomData,
            options: ExportOptions::default(),

            in_heading: false,
            in_drawer: false,
            heading_lvl: 0,
            last_date: "".into(),
            feed_title: "".into(),
            feed_page_id: "".into(),

            filetags: vec![],
            authors: vec![],
        }
    }
}
// First Pass:2 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:3]]
pub fn atomize_file(path: String, options: ExportOptions) -> Result<String> {
    let syntect_handler = SyntectHtmlHandler::new(DefaultHtmlHandler);
    let html_handler = ArroyoHtmlHandler::new(options.clone(), syntect_handler);
    let mut handler = ArroyoAtomHandler::new(options.clone(), html_handler);

    let org = String::from_utf8(fs::read(path.clone())?).unwrap();
    let org_tree = &Org::parse_custom(
        &org,
        &orgize::ParseConfig {
            // Need to pull these from environment or options...
            todo_keywords: (
                vec![
                    "NEXT".to_string(),
                    "INPROGRESS".to_string(),
                    "WAITING".to_string(),
                ],
                vec!["DONE".to_string(), "CANCELLED".to_string()],
            ),
            ..Default::default()
        },
    );

    let mut vec = vec![];
    org_tree.write_html_custom(&mut vec, &mut handler)?;
    Ok(String::from_utf8(vec)?)
}
// First Pass:3 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:4]]
impl<E: From<Error>, H: HtmlHandler<E>> HtmlHandler<E> for ArroyoAtomHandler<E, H> {
    fn start<W: Write>(&mut self, mut w: W, element: &Element) -> Result<(), E> {
        (match element {
// First Pass:4 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:5]]
            Element::Document { .. } => self.start_document(w, element),
// First Pass:5 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:6]]
            Element::Keyword(kw) => self.start_keyword(w, kw),
// First Pass:6 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:7]]
            Element::Title(title) => self.start_title(w, title),
// First Pass:7 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:8]]
            Element::Drawer(drawer) => {
                self.in_drawer = drawer.name == "PROPERTIES" && self.feed_page_id.eq("");
                self.start_rest(w, element)
            }
            Element::Text { value } => self.start_text(w, value),
// First Pass:8 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:9]]
            _t => self.start_rest(w, element),
        })
        .unwrap(); // if we can't parse something, just fucken panic.
// First Pass:9 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:10]]
        Ok(())
    }

    fn end<W: Write>(&mut self, mut w: W, element: &Element) -> Result<(), E> {
        (match element {
// First Pass:10 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:11]]
            // Element::Title(_title) => {}
            Element::Document { .. } => self.end_document(w, element),
// First Pass:11 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:12]]
            Element::Drawer(drawer) => {
                self.in_drawer = false;
                self.end_rest(w, element)
            }
// First Pass:12 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:13]]
            _ => self.end_rest(w, element),
        })
        .ok();
        Ok(())
    }
}
// First Pass:13 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:14]]
impl<E: From<Error>, H: HtmlHandler<E>> ArroyoAtomHandler<E, H> {
// First Pass:14 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:15]]
    fn start_document<W: Write>(&mut self, mut w: W, _document: &elements::Element) -> Result<()> {
        Ok(write!(
            w,
            "<?xml version=\"1.0\" encoding=\"utf-8\"?>
             <feed xmlns=\"http://www.w3.org/2005/Atom\">"
        )?)
    }

    fn end_document<W: Write>(&mut self, mut w: W, _document: &elements::Element) -> Result<()> {
        // the last heading/entry is still "open", close it.
        write!(
            w,
            r#"    </content>
                 </entry>
                 <updated>{}</updated>
               </feed>"#,
            self.last_date,
        )?;
        Ok(())
    }
// First Pass:15 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:16]]
    fn start_title<W: Write>(&mut self, mut w: W, title: &elements::Title) -> Result<()> {
        let ignore_tags = vec![
            String::from("noexport"),
            String::from("NOEXPORT"),
            String::from("ignore"),
        ];
        let export_tags = title
            .tags
            .clone()
            .into_iter()
            .map(String::from)
            .find(|v| ignore_tags.contains(v));
        let props = title.properties.clone().into_hash_map();
        let id = props
            .get("ID")
            .map(|id| id.clone().into())
            .unwrap_or("".to_string());
        let pubdate = props
            .get("PUBDATE")
            .map(|pubdate| pubdate.clone().into())
            .unwrap_or("".to_string());
        let the_link = self.inner.rewrite_link_from(&id);

        if id == "" || pubdate == "" || export_tags.is_some() && self.in_heading {
            self.in_heading = false;
            Ok(())
        } else if id != "" && pubdate != "" && export_tags.is_none() {
            // close previous heading; note that self.heading_lvl defaults to 0
            if title.level <= self.heading_lvl {
                write!(w, "</content>\n")?;
                write!(w, "</entry>\n")?;
            }
            let date = match rfcize_datestamp(pubdate.clone()) {
                Ok(date) => date,
                Err(_) => {
                    dbg!(format!("bad date {}", pubdate.clone()));
                    HtmlEscape(pubdate.clone()).to_string()
                }
            };
            if self.last_date < date {
                self.last_date = date.clone();
            }
            let title_text = match strip_links_from_str(&title.raw.clone()) {
                Ok(text) => HtmlEscape(text),
                Err(the_err) => {
                    dbg!(format!("bad title {} {}", title.raw.clone(), the_err));
                    HtmlEscape(title.raw.to_string())
                }
            };
            let cat_xmls = "";
            let s = format!(
                "<entry>
                   <title>{}</title>
                   <link href=\"{}\"/>
                   <id>{}</id>
                   <updated>{}</updated>\n
                   {}

                   <content type=\"html\">&lt;h{}&gt;", // the HTML encoded heading opening needs to be added here!
                title_text,
                the_link.to_string(), // link
                the_link.to_string(), // ID
                date,
                cat_xmls,
                title.level,
            );
            self.heading_lvl = title.level;
            self.in_heading = true;
            Ok(write!(w, "{}", s)?)
        } else {
            Ok(())
        }
    }
// First Pass:16 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:17]]
    fn start_keyword<W: Write>(&mut self, mut w: W, kw: &elements::Keyword) -> Result<()> {
        // dbg!(kw);
        match kw.key.as_ref() {
            "FILETAGS" => {
                kw.value
                    .split(":")
                    .map(String::from)
                    .filter(|s| !s.is_empty())
                    .for_each(|s| self.filetags.push(s));
                dbg!(&self.filetags);
            }
            "TITLE" => {
                self.feed_title = kw.value.to_string();
                write!(w, r#"<title>{}</title>"#, self.feed_title)?;
            }
            "AUTHOR" => {
                let re = regex::Regex::new(r"(?<name>[\w\s\d]+) <(?<email>.*)>").unwrap();

                re.captures_iter(&kw.value)
                    .map(|caps| {
                        format!(
                            "<author><name>{}</name><email>{}</email></author>",
                            &caps["name"], &caps["email"]
                        )
                    })
                    .for_each(|s| {
                        self.authors.push(s.clone());
                        write!(w, "{}", s).ok();
                    });
                dbg!(&self.authors);
            }
            _ => {}
        }
        Ok(())
    }
// First Pass:17 ends here

// [[file:../arroyo-native-parser.org::*First Pass][First Pass:18]]
    fn start_text<W: Write>(&mut self, mut w: W, text: &Cow<str>) -> Result<()> {
        if self.in_drawer == true {
            let (_, prop_drawer): (_, orgize::elements::PropertiesMap) =
                orgize::elements::Drawer::parse_drawer_content(text)
                    .expect("failed to parse properties drawer");
            let properties = prop_drawer.into_hash_map();

            let mut cur_id: Option<String> = properties.get("ID").map(|s| s.to_string());
            if cur_id.is_none() {
                cur_id = properties.get("CUSTOM_ID").map(|s| s.to_string())
            }
            let the_id = cur_id.expect("Can't publish a page without ID");
            self.feed_page_id = the_id.clone();

            let page_link = self.inner.rewrite_link_from(&the_id);
            write!(
                w,
                r#"<link href="{}"/>
                   <id>{}</id>"#,
                page_link, page_link
            )?;
        } else {
            self.start_rest(
                w,
                &Element::Text {
                    value: text.clone(),
                },
            )?
        }
        Ok(())
    }

    fn start_rest<W: Write>(&mut self, mut w: W, element: &elements::Element) -> Result<()> {
        Ok(if self.in_heading == true {
            let mut buf = InternalWriter::new();
            self.inner.start(&mut buf, element).ok(); // panic if this fails.
            let s = buf.to_utf8().unwrap();
            write!(w, "{}", s)?
        })
    }

    fn end_rest<W: Write>(&mut self, mut w: W, element: &elements::Element) -> Result<()> {
        if self.in_heading == true {
            let mut buf = InternalWriter::new();
            self.inner.end(&mut buf, element).ok();
            let s = buf.to_utf8().unwrap();
            write!(w, "{}", s)?
        }
        Ok(())
    }
}
// First Pass:18 ends here

// [[file:../arroyo-native-parser.org::*Strip Links from Strings][Strip Links from Strings:1]]
fn strip_links_from_str(in_str: &str) -> Result<String> {
    // title.raw.replace("[", "&#91;").replace("]", "&#92;"),
    let re = regex::Regex::new(r"\[(?<wrapped_the_link>\[\])\[(?<text>)\]\]")?;

    Ok(re.replace_all(in_str, "$text").to_string())
}
// Strip Links from Strings:1 ends here

// [[file:../arroyo-native-parser.org::*Convert my org-style timestamps to RFC-3339 strings][Convert my org-style timestamps to RFC-3339 strings:1]]
fn rfcize_datestamp(in_str: String) -> Result<String> {
    let re = regex::Regex::new(
        r"<?(?<year>\d{4})-(?<month>\d{2})-(?<day>\d{2}) \w+ (?<hour>\d{2}):(?<minutes>\d{2})>?",
    )?;
    let date: Option<String> = re
        .captures_iter(&in_str)
        .map(|caps| {
            let year = caps.name("year").unwrap().as_str();
            let month = caps.name("month").unwrap().as_str();
            let day = caps.name("day").unwrap().as_str();
            let hour = caps.name("hour").unwrap().as_str();
            let minutes = caps.name("minutes").unwrap().as_str();
            let ret: String = format!("{}-{}-{}T{}:{}:00-08:00", year, month, day, hour, minutes);
            ret
        })
        .next();
    Ok(date.ok_or(Error::new(std::io::ErrorKind::Other, "invalid date"))?)
}
// Convert my org-style timestamps to RFC-3339 strings:1 ends here

// [[file:../arroyo-native-parser.org::*Internal Buffer Writer for Escaping Entities][Internal Buffer Writer for Escaping Entities:1]]
struct InternalWriter {
    inner: Vec<u8>,
}

impl InternalWriter {
    pub fn to_utf8(self) -> Result<String> {
        return Ok(String::from_utf8(self.inner)?);
    }

    pub fn new() -> InternalWriter {
        return InternalWriter { inner: Vec::new() };
    }
}

impl Write for &mut InternalWriter {
    #[inline]
    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
        let s = String::from_utf8(buf.to_vec()).unwrap();
        let esc = format!("{}", s);
        self.inner.write(esc.as_bytes())
    }

    #[inline]
    fn write_vectored(&mut self, bufs: &[std::io::IoSlice<'_>]) -> std::io::Result<usize> {
        self.inner.write_vectored(bufs)
    }

    #[inline]
    fn flush(&mut self) -> std::io::Result<()> {
        self.inner.flush()
    }

    #[inline]
    fn write_all(&mut self, buf: &[u8]) -> std::io::Result<()> {
        let s = String::from_utf8(buf.to_vec()).unwrap();
        let esc = format!("{}", s);
        // dbg!(s.clone());
        // dbg!(esc.clone());
        self.inner.write_all(esc.as_bytes())
    }

    #[inline]
    fn write_fmt(&mut self, fmt: std::fmt::Arguments<'_>) -> std::io::Result<()> {
        let s = match fmt.as_str() {
            Some(s) => String::from(s),
            _ => fmt.to_string(),
        };

        let s2 = format!("{}", HtmlEscape(s));
        self.inner.write(s2.as_bytes())?;

        Ok(())
    }
}
// Internal Buffer Writer for Escaping Entities:1 ends here
